#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import pytest

from httpx import AsyncClient
from fastapi import HTTPException
from transformers import AutoTokenizer
from vllm.sampling_params import SamplingParams
from vllm.entrypoints.openai import protocol as vllm_protocol

from happy_vllm import utils
from .conftest import TEST_MODELS_DIR
from happy_vllm.routers import functional
from happy_vllm.model.model_base import Model


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def init_model(truncation_side="left"):
    model = Model(app_name=os.environ['app_name'])
    model._tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR, truncation_side=truncation_side)
    model.original_truncation_side = truncation_side
    model.max_model_len = 2048
    return model


def test_verify_request():
    # Without HTTPException
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        temperature=1.0,
        echo=True,
        stream=None
    )
    assert functional.verify_request(request) == None

    # With HTTPException echo and stream
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        echo=True,
        stream=True
    )
    with pytest.raises(HTTPException) as error:
        functional.verify_request(request)
    assert error.value.detail == "Use both echo and stream breaks backend"

    # With HTTPException temperature and top_p equals to 0
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        temperature=0,
        top_p=0
    )
    with pytest.raises(HTTPException) as error:
        functional.verify_request(request)
    assert error.value.detail == "Use temperature and top_p equal to 0 breaks the model"

    # With HTTPException high temperature and top_k equals to 1
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        temperature=2.4,
        top_k=1
    )
    with pytest.raises(HTTPException) as error:
        functional.verify_request(request)
    assert error.value.detail == "Use temperature with high value: 2.4 and top_k equals to 1 : 1 breaks the model"

    # With HTTPException top_p and top_k equals to 1
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        top_p=1,
        top_k=1
    )
    with pytest.raises(HTTPException) as error:
        functional.verify_request(request)
    assert error.value.detail == "Use top_p and top_k equal to 1 breaks the model"

    # With HTTPException max_tokens less than min_tokens
    request = vllm_protocol.ChatCompletionRequest(
        messages=[{"role":"user", "content": "How are you ?"}],
        model="my_model",
        max_tokens=50,
        min_tokens=100
    )
    with pytest.raises(HTTPException) as error:
        functional.verify_request(request)
    assert error.value.detail == "Use max_tokens: 50 less than min_tokens : 100 breaks the model"


@pytest.mark.asyncio
async def test_tokenizer(test_complete_client: AsyncClient):
    """Test the functional route /v1/tokenizer"""
    model = init_model()
    # Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        body = {"text": text, "vanilla": True}
        response = await test_complete_client.post("/tests/v1/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(text)['input_ids']
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        assert set(response_json) == {"tokens_ids", "tokens_nb"}

        # With with_tokens_str
        body = {"text": text, "with_tokens_str": True, "vanilla": True}
        response = await test_complete_client.post("/tests/v1/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(text)['input_ids']
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        tokens_str = model._tokenizer.convert_ids_to_tokens(target_tokens_ids)
        assert response_json["tokens_str"] == tokens_str
        assert set(response_json) == {"tokens_ids", "tokens_nb", "tokens_str"}

    #Non Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        body = {"text": text, "vanilla": False}
        response = await test_complete_client.post("/tests/v1/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = list(utils.proper_tokenization(model._tokenizer, text))
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        assert set(response_json) == {"tokens_ids", "tokens_nb"}

        # With with_tokens_str
        body = {"text": text, "with_tokens_str": True, "vanilla": False}
        response = await test_complete_client.post("/tests/v1/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = list(utils.proper_tokenization(model._tokenizer, text))
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in target_tokens_ids]
        assert response_json["tokens_str"] == tokens_str
        assert set(response_json) == {"tokens_ids", "tokens_nb", "tokens_str"}


@pytest.mark.asyncio
async def test_tokenizer_v2(test_complete_client: AsyncClient):
    """Test the functional route /v2/tokenizer"""
    model = init_model()
    # completions
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        # With add_special_tokens
        body = {"model": "test", "prompt": text, "add_special_tokens": True}
        response = await test_complete_client.post("/tests/v2/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(body["prompt"], add_special_tokens=body['add_special_tokens'])['input_ids']
        assert response_json["tokens"] == target_tokens_ids
        assert response_json["count"] == len(target_tokens_ids)
        assert set(response_json) == {"count", "max_model_len", "tokens"}

        # Without add_special_tokens
        body = {"model": "test", "prompt": text, "add_special_tokens": False}
        response = await test_complete_client.post("/tests/v2/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(body["prompt"], add_special_tokens=body['add_special_tokens'])['input_ids']
        assert response_json["tokens"] == target_tokens_ids
        assert response_json["count"] == len(target_tokens_ids)
        assert set(response_json) == {"count", "max_model_len", "tokens"}


@pytest.mark.asyncio
async def test_decode(test_complete_client: AsyncClient):
    """Test the functional route /v1/decode"""
    model = init_model()

    # Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        token_ids = model._tokenizer(text)['input_ids']

        body = {'token_ids': token_ids, "vanilla": True}
        response = await test_complete_client.post("/tests/v1/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == model._tokenizer.decode(token_ids)
        assert set(response_json) == {"decoded_string"}

        # With with_tokens_str
        body = {'token_ids': token_ids, "vanilla": True, "with_tokens_str": True}
        response = await test_complete_client.post("/tests/v1/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == model._tokenizer.decode(token_ids)
        assert response_json["tokens_str"] == model._tokenizer.convert_ids_to_tokens(token_ids)
        assert set(response_json) == {"decoded_string", "tokens_str"}

    #Non Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        token_ids = model._tokenizer(text)['input_ids']

        body = {'token_ids': token_ids, "vanilla": False}
        response = await test_complete_client.post("/tests/v1/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == utils.proper_decode(model._tokenizer, token_ids)
        assert set(response_json) == {"decoded_string"}

        # With with_tokens_str
        body = {'token_ids': token_ids, "vanilla": False, "with_tokens_str": True}
        response = await test_complete_client.post("/tests/v1/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == utils.proper_decode(model._tokenizer, token_ids)
        assert response_json["tokens_str"] == [utils.proper_decode(model._tokenizer, token_id) for token_id in token_ids]
        assert set(response_json) == {"decoded_string", "tokens_str"}


@pytest.mark.asyncio
async def test_decode_v2(test_complete_client: AsyncClient):
    """Test the functional route /v2/decode"""
    model = init_model()

    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        token_ids = model._tokenizer(text)['input_ids']

        body = {'tokens': token_ids, "model": "test"}
        response = await test_complete_client.post("/tests/v2/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["prompt"] == model._tokenizer.decode(token_ids)
        assert set(response_json) == {"prompt"}


@pytest.mark.asyncio
async def test_split_text(test_complete_client: AsyncClient):
    """Test the route /v1/split_text thanks to the test_complete_client we created in conftest.py"""
    text = "Hey, my name is LLM. How are you ? Fine, you ? That's wonderful news : I'm also fine. But do you think it will last ?"

    body = {"text": text,
            "num_tokens_in_chunk": 2}
    response = await test_complete_client.post("/tests/v1/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ?", 
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 6}
    response = await test_complete_client.post("/tests/v1/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ? Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 1000}
    response = await test_complete_client.post("/tests/v1/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = [text]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 2,
            "separators": [" ?"]}
    response = await test_complete_client.post("/tests/v1/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = target_split_text = ["Hey, my name is LLM. How are you ?",
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine. But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}


@pytest.mark.asyncio
async def test_metadata_text(test_complete_client: AsyncClient):
    """Test the route /v1/metadata_text thanks to the test_complete_client we created in conftest.py"""
    text = "Hey, my name is LLM. How are you ? Fine, and you ? Great."
    model = init_model()
    tokenizer = model._tokenizer

    text_tot = text * 73
    truncation_side = "left"
    max_length = 1234
    body = {"text": text_tot,
            "truncation_side": truncation_side,
            "max_length": max_length}
    response = await test_complete_client.post("/tests/v1/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text

    text_tot = text * 73
    truncation_side = "right"
    max_length = 1234
    body = {"text": text_tot,
            "truncation_side": truncation_side,
            "max_length":1234}
    response = await test_complete_client.post("/tests/v1/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text

    text_tot = text
    truncation_side = "left"
    max_length = 1234
    body = {"text": text,
            "truncation_side": "left",
            "max_length":1234}
    response = await test_complete_client.post("/tests/v1/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import json
import numpy as np
from typing import Any


class NumpyArrayEncoder(json.JSONEncoder):
    """JSONEncoder to store python dict or list containing numpy arrays"""

    def default(self, obj: Any) -> Any:
        """Transform numpy arrays into JSON serializable object such as list
        see : https://docs.python.org/3/library/json.html#json.JSONEncoder.default
        """

        # numpy.ndarray have dtype, astype and tolist attribute and methods that we want
        # to use to convert their element into JSON serializable objects
        if hasattr(obj, "dtype") and hasattr(obj, "astype") and hasattr(obj, "tolist"):

            if np.issubdtype(obj.dtype, np.integer):
                return obj.astype(int).tolist()
            elif np.issubdtype(obj.dtype, np.number):
                return obj.astype(float).tolist()
            else:
                return obj.tolist()

        # sets are not json serializable
        elif isinstance(obj, set):
            return list(obj)

        return json.JSONEncoder.default(self, obj)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import ssl
import json
import torch

from argparse import Namespace, BooleanOptionalAction
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing import Optional, Tuple, Type, Union, List, Mapping, Dict, Any, Literal, get_args

from vllm.utils import FlexibleArgumentParser
from vllm.executor.executor_base import ExecutorBase
from vllm.engine.arg_utils import AsyncEngineArgs, nullable_str
from vllm.entrypoints.openai.tool_parsers import ToolParserManager
from vllm.entrypoints.chat_utils import (ChatTemplateContentFormatOption)
from vllm.entrypoints.openai.reasoning_parsers import ReasoningParserManager
from vllm.entrypoints.openai.cli_args import LoRAParserAction, PromptAdapterParserAction
from vllm.transformers_utils.tokenizer_group.base_tokenizer_group import BaseTokenizerGroup
from vllm.config import (ConfigFormat, TaskOption, HfOverrides, PoolerConfig, CompilationConfig, KVTransferConfig)


DEFAULT_MODEL_NAME = '?'
DEFAULT_EXTRA_INFORMATION = None
DEFAULT_APP_NAME = "happy_vllm"
DEFAULT_API_ENDPOINT_PREFIX = ""
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 5000
DEFAULT_EXPLICIT_ERRORS = False
DEFAULT_ALLOW_CREDENTIALS = False
DEFAULT_ALLOWED_ORIGINS = ["*"]
DEFAULT_ALLOWED_METHODS = ["*"]
DEFAULT_ALLOWED_HEADERS = ["*"]
DEFAULT_UVICORN_LOG_LEVEL = 'info'
CHOICES_UVICORN_LOG_LEVEL = ['debug', 'info', 'warning', 'error', 'critical', 'trace']
DEFAULT_SSL_KEYFILE = None
DEFAULT_SSL_CERTFILE = None
DEFAULT_SSL_CA_CERTS = None
DEFAULT_SSL_CERT_REQS = int(ssl.CERT_NONE)
DEFAULT_ENABLE_SSL_REFRESH = False
DEFAULT_ROOT_PATH = None
DEFAULT_LORA_MODULES = None
DEFAULT_CHAT_TEMPLATE = None
DEFAULT_CHAT_TEMPLATE_CONTENT_FORMAT = "auto"
DEFAULT_RESPONSE_ROLE = "assistant"
DEFAULT_WITH_LAUNCH_ARGUMENTS = False
DEFAULT_MAX_LOG_LEN = None
DEFAULT_PROMPT_ADAPTERS = None
DEFAULT_RETURN_TOKENS_AS_TOKEN_IDS = False
DEFAULT_DISABLE_FRONTEND_MULTIPROCESSING = False
DEFAULT_ENABLE_REQUEST_ID_HEADERS = False
DEFAULT_ENABLE_AUTO_TOOL_CHOICE = False
DEFAULT_TOOL_CALL_PARSER = None
DEFAULT_TOOL_PARSER_PLUGIN = ""
DEFAULT_DISABLE_FASTAPI_DOCS = False
DEFAULT_ENABLE_PROMPT_TOKENS_DETAILS = False
DEFAULT_ENABLE_SERVER_LOAD_TRACKING = False
DEFAULT_DISABLE_UVICORN_ACCESS_LOG = False

class ApplicationSettings(BaseSettings):
    """Application settings

    This class is used for settings management purpose, have a look at the pydantic
    documentation for more details : https://pydantic-docs.helpmanual.io/usage/settings/

    By default, it looks for environment variables (case insensitive) to set the settings
    if a variable is not found, it looks for a file name .env in your working directory
    where you can declare the values of the variables and finally it sets the values
    to the default ones one can define above
    """
    host : str = DEFAULT_HOST
    port: int = DEFAULT_PORT
    explicit_errors: bool = DEFAULT_EXPLICIT_ERRORS
    allow_credentials: bool = DEFAULT_ALLOW_CREDENTIALS
    allowed_origins: list = DEFAULT_ALLOWED_ORIGINS
    allowed_methods: list = DEFAULT_ALLOWED_METHODS
    allowed_headers: list = DEFAULT_ALLOWED_HEADERS
    uvicorn_log_level: str = DEFAULT_UVICORN_LOG_LEVEL
    ssl_keyfile: Optional[str] = DEFAULT_SSL_KEYFILE
    ssl_certfile: Optional[str] = DEFAULT_SSL_CERTFILE
    ssl_ca_certs: Optional[str] = DEFAULT_SSL_CA_CERTS
    ssl_cert_reqs: int = DEFAULT_SSL_CERT_REQS
    enable_ssl_refresh: bool = DEFAULT_ENABLE_SSL_REFRESH
    root_path: Optional[str] = DEFAULT_ROOT_PATH
    app_name: str = DEFAULT_APP_NAME
    api_endpoint_prefix: str = DEFAULT_API_ENDPOINT_PREFIX
    lora_modules: Optional[str] = DEFAULT_LORA_MODULES
    chat_template : Optional[str] = DEFAULT_CHAT_TEMPLATE
    chat_template_content_format: str = DEFAULT_CHAT_TEMPLATE_CONTENT_FORMAT
    response_role: str = DEFAULT_RESPONSE_ROLE
    with_launch_arguments: bool = DEFAULT_WITH_LAUNCH_ARGUMENTS
    max_log_len: Optional[int] = DEFAULT_MAX_LOG_LEN
    prompt_adapters: Optional[str] = DEFAULT_PROMPT_ADAPTERS
    return_tokens_as_token_ids: bool = DEFAULT_RETURN_TOKENS_AS_TOKEN_IDS
    disable_frontend_multiprocessing: bool = DEFAULT_DISABLE_FRONTEND_MULTIPROCESSING
    enable_request_id_headers: bool = DEFAULT_ENABLE_REQUEST_ID_HEADERS
    enable_auto_tool_choice: bool = DEFAULT_ENABLE_AUTO_TOOL_CHOICE
    tool_call_parser: Optional[str] = DEFAULT_TOOL_CALL_PARSER
    tool_parser_plugin: Optional[str] = DEFAULT_TOOL_PARSER_PLUGIN
    disable_fastapi_docs : Optional[bool] = DEFAULT_DISABLE_FASTAPI_DOCS
    enable_prompt_tokens_details : Optional[bool] = DEFAULT_ENABLE_PROMPT_TOKENS_DETAILS
    enable_server_load_tracking: Optional[bool]= DEFAULT_ENABLE_SERVER_LOAD_TRACKING
    disable_uvicorn_access_log: bool = DEFAULT_DISABLE_UVICORN_ACCESS_LOG


    model_config = SettingsConfigDict(env_file=".env", extra='ignore', protected_namespaces=('settings', ))


def get_model_settings(parser: FlexibleArgumentParser) -> BaseSettings:
    """Gets the model settings. It corresponds to the variables added via AsyncEngineArgs.add_cli_args plus model-name and extra-information.
    First we use the parser to get the default values of vLLM for these variables. We instantiate a BaseSettings model
    with these values as default. They are possibly overwritten by environnement variables or those of a .env

    Args:
        parser (FlexibleArgumentParser) : The parser containing all the model variables with their default values from vLLM 
    """

    default_args = parser.parse_args([])
    # Define ModelSettings with the default values of the args (which will be replaced)
    # by the environnement variables or those of the .env file
    class ModelSettings(BaseSettings):
        model: str = default_args.model
        model_name: str = default_args.model_name
        extra_information: Optional[str] = default_args.extra_information
        served_model_name: Optional[Union[str, List[str]]] = None
        tokenizer: Optional[str] = default_args.tokenizer
        hf_config_path: Optional[str] = default_args.hf_config_path
        task: TaskOption = default_args.task
        skip_tokenizer_init: bool = False
        tokenizer_mode: str = default_args.tokenizer_mode
        trust_remote_code: bool = False

        allowed_local_media_path: Optional[str] = default_args.allowed_local_media_path
        download_dir: Optional[str] = default_args.download_dir
        load_format: str = default_args.load_format
        config_format: ConfigFormat = default_args.config_format
        dtype: str = default_args.dtype
        kv_cache_dtype: str = default_args.kv_cache_dtype
        seed: Optional[int] = default_args.seed
        max_model_len: Optional[int] = default_args.max_model_len
        distributed_executor_backend: Optional[Union[str, ExecutorBase]] = default_args.distributed_executor_backend
        pipeline_parallel_size: int = default_args.pipeline_parallel_size
        tensor_parallel_size: int = default_args.tensor_parallel_size
        enable_expert_parallel: bool = False
        max_parallel_loading_workers: Optional[int] = default_args.max_parallel_loading_workers
        block_size: Optional[int] = default_args.block_size
        enable_prefix_caching: Optional[bool] = default_args.enable_prefix_caching
        disable_sliding_window: bool = False
        disable_cascade_attn: bool = False
        swap_space: float = default_args.swap_space # GiB
        cpu_offload_gb: float = default_args.cpu_offload_gb  # GiB
        gpu_memory_utilization: float = default_args.gpu_memory_utilization
        max_num_batched_tokens: Optional[int] = default_args.max_num_batched_tokens
        max_num_partial_prefills: Optional[int] = default_args.max_num_partial_prefills
        max_long_partial_prefills: Optional[int] = default_args.max_long_partial_prefills
        long_prefill_token_threshold: Optional[int] = default_args.long_prefill_token_threshold
        max_num_seqs: Optional[int] = default_args.max_num_seqs
        disable_log_stats: bool = False
        revision: Optional[str] = default_args.revision
        code_revision: Optional[str] = default_args.code_revision
        rope_scaling: Optional[Dict[str, Any]] = default_args.rope_scaling
        rope_theta: Optional[float] = None
        hf_overrides: Optional[HfOverrides] = default_args.hf_overrides
        tokenizer_revision: Optional[str] = default_args.tokenizer_revision
        quantization: Optional[str] = default_args.quantization
        enforce_eager: Optional[bool] = default_args.enforce_eager
        max_seq_len_to_capture: int = default_args.max_seq_len_to_capture
        disable_custom_all_reduce: bool = False
        enable_lora: bool = False
        enable_lora_bias: bool = False
        max_loras: int = default_args.max_loras
        max_lora_rank: int = default_args.max_lora_rank
        enable_prompt_adapter: bool = False
        max_prompt_adapters: int = default_args.max_prompt_adapters
        max_prompt_adapter_token: int = default_args.max_prompt_adapter_token
        fully_sharded_loras: bool = False
        lora_extra_vocab_size: int = default_args.lora_extra_vocab_size
        long_lora_scaling_factors: Optional[Tuple[float]] = default_args.long_lora_scaling_factors
        lora_dtype: Optional[Union[str, torch.dtype]] = default_args.lora_dtype
        max_cpu_loras: Optional[int] = default_args.max_cpu_loras
        device: str = default_args.device
        num_scheduler_steps: int = default_args.num_scheduler_steps
        multi_step_stream_outputs: bool = default_args.multi_step_stream_outputs
        ray_workers_use_nsight: bool = False
        num_gpu_blocks_override: Optional[int] = default_args.num_gpu_blocks_override
        num_lookahead_slots: int = default_args.num_lookahead_slots
        model_loader_extra_config: Optional[dict] = default_args.model_loader_extra_config
        ignore_patterns:  Optional[Union[str, List[str]]] = default_args.ignore_patterns
        preemption_mode: Optional[str] = default_args.preemption_mode
        disable_log_requests: bool = False
        engine_use_ray: bool = False
        use_v2_block_manager: bool = default_args.use_v2_block_manager
        max_logprobs: int = default_args.max_logprobs
        tokenizer_pool_size: int = default_args.tokenizer_pool_size
        tokenizer_pool_type: Union[str, BaseTokenizerGroup] = default_args.tokenizer_pool_type
        tokenizer_pool_extra_config: Optional[Dict[str, Any]] = default_args.tokenizer_pool_extra_config
        limit_mm_per_prompt: Optional[Mapping[str, int]] = default_args.limit_mm_per_prompt
        mm_processor_kwargs: Optional[Dict[str, Any]] = default_args.mm_processor_kwargs
        disable_mm_preprocessor_cache: bool = False
        scheduling_policy: Literal["fcfs", "priority"] = default_args.scheduling_policy
        scheduler_cls: Union[str, Type[object]] = default_args.scheduler_cls
        scheduler_delay_factor: float = default_args.scheduler_delay_factor
        enable_chunked_prefill: Optional[bool] = default_args.enable_chunked_prefill
        guided_decoding_backend: str = default_args.guided_decoding_backend
        logits_processor_pattern: Optional[str] = default_args.logits_processor_pattern
        speculative_config: Optional[Union[str, Dict[str, Any]]] = default_args.speculative_config
        speculative_model: Optional[str] = default_args.speculative_model
        speculative_model_quantization: Optional[str] = default_args.speculative_model_quantization
        speculative_draft_tensor_parallel_size: Optional[int] = default_args.speculative_draft_tensor_parallel_size
        num_speculative_tokens: Optional[int] = default_args.num_speculative_tokens
        speculative_disable_mqa_scorer: Optional[bool] = default_args.speculative_disable_mqa_scorer
        speculative_max_model_len: Optional[int] = default_args.speculative_max_model_len
        speculative_disable_by_batch_size: Optional[int] = default_args.speculative_disable_by_batch_size
        ngram_prompt_lookup_max: Optional[int] = default_args.ngram_prompt_lookup_max
        ngram_prompt_lookup_min: Optional[int] = default_args.ngram_prompt_lookup_min
        spec_decoding_acceptance_method: str = default_args.spec_decoding_acceptance_method
        typical_acceptance_sampler_posterior_threshold: Optional[float] = default_args.typical_acceptance_sampler_posterior_threshold
        typical_acceptance_sampler_posterior_alpha: Optional[float] = default_args.typical_acceptance_sampler_posterior_alpha
        disable_logprobs_during_spec_decoding: Optional[bool] = default_args.disable_logprobs_during_spec_decoding
        qlora_adapter_name_or_path: Optional[str] = default_args.qlora_adapter_name_or_path
        show_hidden_metrics_for_version: Optional[str] = default_args.show_hidden_metrics_for_version
        disable_async_output_proc: bool = False
        override_neuron_config: Optional[Dict[str, Any]] = default_args.override_neuron_config
        override_pooler_config: Optional[PoolerConfig] = default_args.override_pooler_config
        compilation_config: Optional[CompilationConfig] = default_args.compilation_config
        worker_cls: str = default_args.worker_cls
        worker_extension_cls: str = default_args.worker_extension_cls
        kv_transfer_config: Optional[KVTransferConfig] = default_args.kv_transfer_config
        generation_config: Optional[str] = default_args.generation_config
        override_generation_config: Optional[Dict[str, Any]] = default_args.override_generation_config
        enable_sleep_mode: bool = False
        model_impl: str = default_args.model_impl

        calculate_kv_scales: Optional[bool] = default_args.calculate_kv_scales
        additional_config: Optional[Dict[str, Any]] = default_args.additional_config
        enable_reasoning: Optional[bool] = default_args.enable_reasoning
        reasoning_parser: Optional[str] = default_args.reasoning_parser
        use_tqdm_on_load: bool = default_args.use_tqdm_on_load
        otlp_traces_endpoint: Optional[str] = default_args.otlp_traces_endpoint
        collect_detailed_traces: Optional[str] = default_args.collect_detailed_traces

        model_config = SettingsConfigDict(env_file=".env", extra='ignore', protected_namespaces=('settings', ))

    model_settings = ModelSettings()
    return model_settings


def get_parser() -> FlexibleArgumentParser:
    """Gets the parser. The default values of all application variables (see ApplicationSettings) are properly
    set to the BaseSetting value defined via pydantic. The default values of all model variables (ie those added
    via AsyncEngineArgs.add_cli_args plus model-name) are not properly set via pydantic at this point.
    
    Returns:
        FlexibleArgumentParser : The argparse parser
    """
    parser = FlexibleArgumentParser(description="REST API server for vLLM, production ready")

    application_settings = ApplicationSettings(_env_parse_none_str='None') # type: ignore

    parser.add_argument("--host",
                        type=str,
                        default=application_settings.host,
                        help="host name")
    parser.add_argument("--port",
                        type=int,
                        default=application_settings.port,
                        help="port number")
    parser.add_argument("--model-name",
                        type=str,
                        default=DEFAULT_MODEL_NAME,
                        help="The name of the model given by the /info endpoint of the API")
    parser.add_argument("--extra-information",
                        type=str,
                        default=DEFAULT_EXTRA_INFORMATION,
                        help="The path to a json to add to the /info endpoint of the API")
    parser.add_argument("--app-name",
                        type=str,
                        default=application_settings.app_name,
                        help="The name of the application")
    parser.add_argument("--api-endpoint-prefix",
                        type=str,
                        default=application_settings.api_endpoint_prefix,
                        help="The prefix for the API endpoints")
    parser.add_argument("--explicit-errors",
                        default=application_settings.explicit_errors,
                        action=BooleanOptionalAction,
                        help="If True, the underlying python errors are sent back via the API")
    parser.add_argument('--allow-credentials',
                        default=application_settings.allow_credentials,
                        action=BooleanOptionalAction,
                        help="allow credentials for CORS")
    parser.add_argument("--allowed-origins",
                        type=json.loads,
                        default=application_settings.allowed_origins,
                        help="CORS allowed origins")
    parser.add_argument("--allowed-methods",
                        type=json.loads,
                        default=application_settings.allowed_methods,
                        help="CORS allowed methods")
    parser.add_argument("--allowed-headers",
                        type=json.loads,
                        default=application_settings.allowed_headers,
                        help="CORS allowed headers")
    parser.add_argument("--uvicorn-log-level",
                        type=str,
                        default=application_settings.uvicorn_log_level,
                        choices=CHOICES_UVICORN_LOG_LEVEL,
                        help="log level for uvicorn")
    parser.add_argument("--ssl-keyfile",
                        type=str,
                        default=application_settings.ssl_keyfile,
                        help="The file path to the SSL key file")
    parser.add_argument("--ssl-certfile",
                        type=str,
                        default=application_settings.ssl_certfile,
                        help="The file path to the SSL cert file")
    parser.add_argument("--ssl-ca-certs",
                        type=str,
                        default=application_settings.ssl_ca_certs,
                        help="The CA certificates file")
    parser.add_argument("--enable-ssl-refresh",
                        action="store_true",
                        default=application_settings.enable_ssl_refresh,
                        help="Refresh SSL Context when SSL certificate files change")
    parser.add_argument("--ssl-cert-reqs",
                        type=int,
                        default=application_settings.ssl_cert_reqs,
                        help="Whether client certificate is required (see stdlib ssl module's)")
    parser.add_argument("--root-path",
                        type=str,
                        default=application_settings.root_path,
                        help="FastAPI root_path when app is behind a path based routing proxy")
    parser.add_argument("--lora-modules",
                        type=str,
                        default=application_settings.lora_modules,
                        nargs='+',
                        action=LoRAParserAction,
                        help="LoRA module configurations in the format name=path. "
                        "Multiple modules can be specified.")
    parser.add_argument("--chat-template",
                        type=str,
                        default=application_settings.chat_template,
                        help="The file path to the chat template, "
                        "or the template in single-line form "
                        "for the specified model")
    parser.add_argument('--chat-template-content-format',
                        type=str,
                        default=application_settings.chat_template_content_format,
                        choices=get_args(ChatTemplateContentFormatOption),
                        help='The format to render message content within a chat template.'
                        '\n\n'
                        '* "string" will render the content as a string. '
                        'Example: "Hello World"\n'
                        '* "openai" will render the content as a list of dictionaries, '
                        'similar to OpenAI schema. '
                        'Example: [{"type": "text", "text": "Hello world!"}]')
    parser.add_argument("--response-role",
                        type=str,
                        default=application_settings.response_role,
                        help="The role name to return if "
                        "`request.add_generation_prompt=true`.")
    parser.add_argument("--with-launch-arguments",
                        type=bool,
                        default=application_settings.with_launch_arguments,
                        help="Whether the route launch_arguments should display the launch arguments")
    parser.add_argument("--return-tokens-as-token-ids",
                        default=application_settings.return_tokens_as_token_ids,
                        action=BooleanOptionalAction,
                        help="When --max-logprobs is specified, represents single tokens as "
                        "strings of the form 'token_id:{token_id}' so that tokens that "
                        "are not JSON-encodable can be identified.")
    parser.add_argument('--max-log-len',
                        type=int,
                        default=application_settings.max_log_len,
                        help='Max number of prompt characters or prompt '
                        'ID numbers being printed in log.'
                        '\n\nDefault: Unlimited')
    parser.add_argument("--prompt-adapters",
                        type=nullable_str,
                        default=application_settings.prompt_adapters,
                        nargs='+',
                        action=PromptAdapterParserAction,
                        help="Prompt adapter configurations in the format name=path. "
                        "Multiple adapters can be specified.")
    parser.add_argument("--disable-frontend-multiprocessing",
                        default=application_settings.disable_frontend_multiprocessing,
                        action=BooleanOptionalAction,
                        help="If specified, will run the OpenAI frontend server in the same "
                        "process as the model serving engine.")
    parser.add_argument("--enable-request-id-headers",
                        default=application_settings.enable_request_id_headers,
                        action=BooleanOptionalAction,
                        help="If specified, API server will add X-Request-Id header to "
                        "responses. Caution: this hurts performance at high QPS.")
    parser.add_argument("--enable-auto-tool-choice",
                        default=application_settings.enable_auto_tool_choice,
                        action=BooleanOptionalAction,
                        help=
                        "Enable auto tool choice for supported models. Use --tool-call-parser"
                        "to specify which parser to use")
    valid_reasoning_parsers = ReasoningParserManager.reasoning_parsers.keys()
    valid_tool_parsers = ToolParserManager.tool_parsers.keys()
    parser.add_argument("--tool-call-parser",
                        type=str,
                        metavar="{" + ",".join(valid_tool_parsers) + "} or name registered in "
                        "--tool-parser-plugin",
                        default=application_settings.tool_call_parser,
                        help=
                        "Select the tool call parser depending on the model that you're using."
                        " This is used to parse the model-generated tool call into OpenAI API "
                        "format. Required for --enable-auto-tool-choice.")
    parser.add_argument("--tool-parser-plugin",
                        type=str,
                        default=application_settings.tool_parser_plugin,
                        help=
                        "Special the tool parser plugin write to parse the model-generated tool"
                        " into OpenAI API format, the name register in this plugin can be used "
                        "in --tool-call-parser.")
    parser.add_argument("--disable-fastapi-docs",
                        action='store_true',
                        default=application_settings.disable_fastapi_docs,
                        help="Disable FastAPI's OpenAPI schema, Swagger UI, and ReDoc endpoint")
    parser.add_argument("--enable-prompt-tokens-details",
                        action='store_true',
                        default=application_settings.enable_prompt_tokens_details,
                        help="If set to True, enable prompt_tokens_details in usage.")
    parser.add_argument("--enable-server-load-tracking",
                        action='store_true',
                        default=application_settings.enable_server_load_tracking,
                        help="If set to True, enable tracking server_load_metrics in the app state.")
    parser.add_argument("--disable-uvicorn-access-log",
                        action="store_true",
                        help="Disable uvicorn access log.")

    parser = AsyncEngineArgs.add_cli_args(parser)
    return parser


def parse_args() -> Namespace:
    """Parses the args. We want this priority : args from cli > environnement variables > .env variables.
    In order to do this, we get from pydantic BaseSetting the value from environnement variables or .env variables
    with the proper priority. Then we set the default value of the cli parser to those value so that if the cli args
    are used, they overwrite the default values and otherwise, the BaseSetting value is taken.

    Returns:
        NameSpace
    """
    # Gets the parser
    # The default value of the application variables are properly set
    # Those of the model variables are not
    parser = get_parser()
    # Gets the default values of the model variables via pydantic
    model_settings = get_model_settings(parser)
    # Sets the default values of the model variables in the parser
    parser.set_defaults(**model_settings.model_dump())
    # Gets the args
    args = parser.parse_args()
    # Explicitly check for help flag for the providing help message to the entrypoint
    if '-h' in sys.argv[1:] or '--help' in sys.argv[1:]:
        parser.print_help()
        sys.exit()
    return args



#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Resources for the FastAPI application

This module define resources that need to be instantiated at startup in a global
variable resources that can be used in routes.

This is the way your machine learning models can be loaded in memory at startup
so they can handle requests.
"""

import logging
from typing import Callable
from argparse import Namespace
from contextlib import asynccontextmanager

from fastapi import FastAPI
from ..model.model_base import Model

from vllm.engine.multiprocessing.client import MQLLMEngineClient



logger = logging.getLogger(__name__)

RESOURCES = {}
RESOURCE_MODEL = "model"

def get_lifespan(async_engine_client: MQLLMEngineClient, args: Namespace) -> Callable:
    @asynccontextmanager
    async def lifespan(app: FastAPI):
        # Load the ML model
        model = Model(app_name=args.app_name)
        await model.loading(async_engine_client, args=args)
        logger.info("Model loaded")

        RESOURCES[RESOURCE_MODEL] = model

        # Force log once to have informations in /metrics before requests
        await model._model.do_log_stats()

        yield

        # Clean up the ML models and release the resources
        RESOURCES.clear()
    return lifespan



#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import json
import shutil
import pytest
from httpx import AsyncClient

from .conftest import TEST_MODELS_DIR, TEST_DATA_DIR


from happy_vllm import utils


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)

@pytest.mark.asyncio
async def test_get_liveness(test_base_client: AsyncClient):
    """Test the technical route /liveness"""
   
    response = await test_base_client.get("/tests/liveness")
    assert response.status_code == 200
    assert response.json() == {"alive": "ok"}


@pytest.mark.asyncio
async def test_get_readiness_ko(test_base_client: AsyncClient):
    """Test the technical route /readiness when the model is not loaded"""
    response = await test_base_client.get("/tests/readiness")
    assert response.status_code == 200
    assert response.json() == {"ready": "ko"}


@pytest.mark.asyncio
async def test_get_readiness_ok(test_complete_client: AsyncClient):
    """Test the technical route /readiness when the model is fully loaded"""
    response = await test_complete_client.get("/tests/readiness")
    assert response.status_code == 200
    assert response.json() == {"ready": "ok"}


@pytest.mark.asyncio
async def test_info(test_complete_client: AsyncClient):
    """Test the technical route /v1/info"""
    response = await test_complete_client.get("/tests/v1/info")
    assert response.status_code == 200
    response_json = response.json()
    assert response_json["application"] == "APP_TESTS"
    assert response_json["version"] == utils.get_package_version()
    assert response_json["model_name"] == "TEST MODEL"
    assert response_json["truncation_side"] == "right"
    assert response_json["max_length"] == 2048
    with open(TEST_DATA_DIR / "extra_information.json", "r") as json_file:
        target_extra_information = json.load(json_file)
    assert set(response_json["extra_information"]) == set(target_extra_information)
    for key, value in target_extra_information.items():
        response_json["extra_information"][key] == value


@pytest.mark.asyncio
async def test_launch_arguments(test_complete_client: AsyncClient):
    """Test the technical route /v1/launch_arguments"""
    response = await test_complete_client.get("/tests/v1/launch_arguments")
    assert response.status_code == 200
    response_json = response.json()
    assert response_json["app_name"] == "APP_TESTS"
    assert response_json["model_name"] == "TEST MODEL"
    assert response_json["with_launch_arguments"] == True


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""This module contains the base Model class
"""


import os
import json
import asyncio
import logging
from pathlib import Path
from argparse import Namespace
from transformers import AutoTokenizer
from typing import Any, Tuple, Union, List, cast
from vllm.entrypoints.logger import RequestLogger
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.entrypoints.chat_utils import load_chat_template
from vllm.engine.multiprocessing.client import MQLLMEngineClient
from vllm.entrypoints.openai.serving_chat import OpenAIServingChat
from vllm.entrypoints.openai.serving_embedding import OpenAIServingEmbedding
from vllm.entrypoints.openai.serving_completion import OpenAIServingCompletion
from vllm.entrypoints.openai.protocol import TokenizeResponse, DetokenizeResponse
from vllm.entrypoints.openai.serving_tokenization import OpenAIServingTokenization
from vllm.transformers_utils.tokenizer_group.tokenizer_group import TokenizerGroup
from vllm.entrypoints.openai.serving_transcription import OpenAIServingTranscription
from vllm.entrypoints.openai.serving_models import (BaseModelPath, OpenAIServingModels)

from happy_vllm import utils


logger = logging.getLogger(__name__)


class Model:
    """Parent model class.
    """

    def __init__(self, **kwargs):
        '''Init. model class'''
        self._model = None
        self._tokenizer = None
        self._model_conf = None
        self._model_explainer = None
        self.openai_serving_chat = None
        self.openai_serving_embedding = None
        self.openai_serving_completion = None
        self.openai_serving_tokenization = None
        self._loaded = False
        self.app_name = kwargs.get('app_name', "happy_vllm")
        self.extra_information = {}

    def is_model_loaded(self):
        """return the state of the model"""
        return self._loaded

    async def loading(self, async_engine_client: MQLLMEngineClient, args: Namespace, **kwargs):
        """load the model"""
        await self._load_model(async_engine_client, args, **kwargs)
        self._loaded = True
        if args.extra_information:
            with open(args.extra_information, 'r') as json_file:
                self.extra_information = json.load(json_file)
        if args.with_launch_arguments:
            self.launch_arguments = vars(args)
        else:
            self.launch_arguments = {}

    async def _load_model(self, async_engine_client: MQLLMEngineClient, args: Namespace, **kwargs) -> None:
        """Load a model from a file

        Returns:
            Tuple[Any, dict]: A tuple containing the model and a dict of metadata about it.
        """

        self._model_conf = {'model_name': args.model_name}
    
        logger.info(f"Loading the model from {args.model}")
        if args.model_name != "TEST MODEL":
            self._model = async_engine_client
            model_config = await self._model.get_model_config()
            # Define the tokenizer differently if we have an AsyncLLMEngine
            if isinstance(self._model, AsyncLLMEngine):
                tokenizer_tmp = self._model.engine.tokenizer
            else:
                tokenizer_tmp = self._model.tokenizer
            if isinstance(tokenizer_tmp, TokenizerGroup): # type: ignore
                self._tokenizer = tokenizer_tmp.tokenizer # type: ignore
            else:
                self._tokenizer = tokenizer_tmp # type: ignore
            self.max_model_len = model_config.max_model_len # type: ignore
            # To take into account Mistral tokenizers
            try:
                self.original_truncation_side = self._tokenizer.truncation_side # type: ignore
            except:
                self.original_truncation_side = "left"
            if args.disable_log_requests:
                request_logger = None
            else:
                request_logger = RequestLogger(max_log_len=args.max_log_len)
            served_model_names = [args.model_name]
            base_model_paths = [
                BaseModelPath(name=name, model_path=args.model)
                for name in served_model_names
            ]
            resolved_chat_template = load_chat_template(args.chat_template)
            logger.info("Using supplied chat template:\n%s", resolved_chat_template)
            self.openai_serving_models = OpenAIServingModels(
                                            engine_client=cast(AsyncLLMEngine,self._model),
                                            model_config=model_config,
                                            base_model_paths=base_model_paths,
                                            lora_modules=args.lora_modules,
                                            prompt_adapters=args.prompt_adapters,
                                        )
            self.task = model_config.task
            if model_config.task == "generate":
                self.openai_serving_chat = OpenAIServingChat(cast(AsyncLLMEngine,self._model), model_config,
                                                            self.openai_serving_models,
                                                            args.response_role,
                                                            request_logger=request_logger,
                                                            chat_template=resolved_chat_template,
                                                            chat_template_content_format=args.chat_template_content_format,
                                                            return_tokens_as_token_ids=args.return_tokens_as_token_ids,
                                                            enable_auto_tools=args.enable_auto_tool_choice,
                                                            tool_parser=args.tool_call_parser,
                                                            enable_reasoning=args.enable_reasoning,
                                                            reasoning_parser=args.reasoning_parser,
                                                            enable_prompt_tokens_details=args.enable_prompt_tokens_details)
                self.openai_serving_completion = OpenAIServingCompletion(cast(AsyncLLMEngine,self._model), model_config, 
                                                                        self.openai_serving_models, 
                                                                        request_logger=request_logger,
                                                                        return_tokens_as_token_ids=args.return_tokens_as_token_ids)
            if model_config.task == "embed":
                self.openai_serving_embedding = OpenAIServingEmbedding(cast(AsyncLLMEngine,self._model),
                                                                        model_config,
                                                                        self.openai_serving_models,
                                                                        request_logger=request_logger,
                                                                        chat_template=resolved_chat_template,
                                                                        chat_template_content_format=args.chat_template_content_format)
            if model_config.task == "transcription":
                self.openai_serving_transcription = OpenAIServingTranscription(cast(AsyncLLMEngine,self._model),
                                                                                model_config,
                                                                                self.openai_serving_models,
                                                                                request_logger=request_logger)
            self.openai_serving_tokenization = OpenAIServingTokenization(cast(AsyncLLMEngine,self._model), model_config, 
                                                                        self.openai_serving_models,
                                                                        request_logger=request_logger,
                                                                        chat_template=resolved_chat_template,
                                                                        chat_template_content_format=args.chat_template_content_format)

        # For test purpose
        else:
            self.max_model_len = 2048
            self.original_truncation_side = 'right'
            self._tokenizer = AutoTokenizer.from_pretrained(utils.TEST_TOKENIZER_NAME,
                                                     cache_dir=os.environ["TEST_MODELS_DIR"], truncation_side=self.original_truncation_side)
            self._model = MockModel(self._tokenizer, self.app_name)
            self.openai_serving_tokenization = MockOpenAIServingTokenization(self._tokenizer)
        logger.info(f"Model loaded")

    def tokenize(self, text: str) -> List[int]:
        """Tokenizes a text
        
        Args:
            text (str) : The text to tokenize

        Returns:
            list : The list of token ids
        """
        return list(utils.proper_tokenization(self._tokenizer, text))

    def split_text(self, text: str, num_tokens_in_chunk: int = 200, separators: Union[list, None] = None) -> List[str]:
        '''Splits a text in small texts containing at least num_tokens_in_chunk tokens and ending by a separator. note that the `separators` 
        used are the tokenization of the strings and not the strings themselves (which explains why we must for example 
        specify ' .' and '.' as two separate separators) 

        Args:
            text (str) : The text to split

        Kwargs:
            num_tokens_in_chunk (int) : The minimal number of tokens in the chunk
            separators (list) : The separators marking the end of a sentence

        Returns:
            A list of strings each string containing at least num_tokens_in_chunk tokens and ending by a separator
        '''
        if separators is None:
            separators = [".", "!", "?", "|", " .", " !", " ?", " |"]
        separators_tokens_ids = set()
        for separator in separators:
            separators_tokens_ids.add(utils.proper_tokenization(self._tokenizer, separator))
        tokens = list(utils.proper_tokenization(self._tokenizer, text))
        indices_separators = []
        for separator_tokens_ids in separators_tokens_ids:
            indices_separators += find_indices_sub_list_in_list(tokens, list(separator_tokens_ids))
        indices_separators.sort()

        chunks = []
        index_beginning_chunk = 0
        current_used_separator = 0
        while current_used_separator < len(indices_separators):
            index_current_used_separator = indices_separators[current_used_separator]
            if index_current_used_separator +1 - index_beginning_chunk >= num_tokens_in_chunk:
                chunks.append(tokens[index_beginning_chunk:index_current_used_separator + 1])
                index_beginning_chunk = index_current_used_separator + 1
            current_used_separator += 1
        chunks.append(tokens[index_beginning_chunk:])
        chunks_decoded = [utils.proper_decode(self._tokenizer, chunk) for chunk in chunks]
        chunks_decoded = [element for element in chunks_decoded if element!= ""]
        return chunks_decoded

    def extract_text_outside_truncation(self, text: str, truncation_side: Union[str, None] = None, max_length: Union[int, None] = None) -> str:
        """Extracts the part of the prompt not kept after truncation, which will not be infered by the model.
        First, we tokenize the prompt while applying truncation.
        We obtain a list of sequences of token ids padded, which are outside the truncation.
        Then we decode this list of tensors of token IDs containing special tokens to a string.

        Args:
            text (str) : The text we want to parse
            truncation_side (str) : The side of the truncation
            max_length (int) : The length above which the text will be truncated

        Returns:
            The part of the text which will be dropped by the truncation (str)
        """
        if max_length is None:
            max_length = self.max_model_len
        if truncation_side is None:
            truncation_side = self.original_truncation_side
        self._tokenizer.truncation_side = truncation_side
        list_tokens = self._tokenizer(text, truncation=True, add_special_tokens=False, max_length=max_length, return_overflowing_tokens=True)['input_ids']
        if len(list_tokens) <= 1:
            return ''
        not_truncated = list_tokens[0]
        truncated_tmp = list_tokens[1:]
        if self._tokenizer.truncation_side == 'left':
            truncated_tmp.reverse()
        truncated = []
        for truncated_tokens in truncated_tmp:
            truncated += truncated_tokens
        truncated_str = self._tokenizer.decode(truncated)
        self._tokenizer.truncation_side = self.original_truncation_side
        return truncated_str


def find_indices_sub_list_in_list(big_list: list, sub_list: list) -> list:
    """Find the indices of the presence of a sub list in a bigger list. For example
    if big_list = [3, 4, 1, 2, 3, 4, 5, 6, 3, 4] and sub_list = [3, 4],
    the result will be [1, 5, 9]

    Args:
        big_list (list) : The list in which we want to find the sub_list
        sub_list (list): The list we want the indices of in the big_list

    Returns:
        list : The list of indices of where the sub_list is in the big_list 
    """
    len_sub_list = len(sub_list)
    indices = []
    for index in range(len(big_list)):
        if big_list[index - len_sub_list + 1: index + 1] == sub_list:
            indices.append(index)
    return indices


class MockOpenAIServingTokenization():

    def __init__(self, tokenizer):
        self.tokenizer=tokenizer

    async def create_tokenize(self, request, raw_request):
        token = self.tokenizer(request.prompt, add_special_tokens=request.add_special_tokens)['input_ids']
        return TokenizeResponse(tokens=token,
                                count=len(token),
                                max_model_len=1)

    async def create_detokenize(self, request, raw_request):
        decode = self.tokenizer.decode(request.tokens)
        return DetokenizeResponse(prompt=decode)


class MockModel():

    def __init__(self, tokenizer, app_name: str = "happy_vllm"):
        self.tokenizer = tokenizer
        self.app_name = app_name

    async def async_iter(self, my_list):
        for element in my_list:
            yield element

    async def do_log_stats(self):
        pass


class MockGenerateResponse():

    def __init__(self, prompt, outputs, tokenizer):
        self.prompt = prompt
        outputs = [MockOutput(output, tokenizer) for output in outputs]
        self.outputs = outputs
        self.prompt_token_ids = tokenizer(self.prompt)['input_ids']


class MockOutput():
    
    def __init__(self, text, tokenizer):
        self.text = text
        self.token_ids = tokenizer(text)['input_ids']
        if text[-len("don't you ?"):] == "don't you ?":
            self.finish_reason = "stop"
        else:
            self.finish_reason = None





#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import json
import pytest
import shutil
from argparse import Namespace
from transformers import AutoTokenizer

from happy_vllm import utils
from happy_vllm.model import model_base
from happy_vllm.model.model_base import Model
from happy_vllm.launch import happy_vllm_build_async_engine_client

from .conftest import TEST_MODELS_DIR, TEST_DATA_DIR


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def init_model(truncation_side="left"):
    model = Model(app_name=os.environ['app_name'])
    model._tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR, truncation_side=truncation_side)
    model.original_truncation_side = truncation_side
    model.max_model_len = 2048
    return model


@pytest.mark.asyncio
async def test_is_model_loaded():
    args = Namespace(model_name=os.environ["MODEL_NAME"], model=os.environ['MODEL'], with_launch_arguments=True, extra_information=None)
    model = init_model()
    assert not(model.is_model_loaded())
    await model.loading(happy_vllm_build_async_engine_client(args), args)
    assert model.is_model_loaded()


@pytest.mark.asyncio
async def test_loading():
    # Nominal case
    args = Namespace(model_name=os.environ["MODEL_NAME"], model=os.environ['MODEL'], with_launch_arguments=False, extra_information=None)
    model = init_model()
    await model.loading(happy_vllm_build_async_engine_client(args), args)
    assert model.launch_arguments == {}
    assert model.extra_information == {}

    # with_launch_arguments True
    args = Namespace(model_name=os.environ["MODEL_NAME"], model=os.environ['MODEL'], with_launch_arguments=True, extra_information=None)
    model = init_model()
    await model.loading(happy_vllm_build_async_engine_client(args), args)
    assert model.extra_information == {}
    target_launch_arguments = vars(args)
    assert set(target_launch_arguments) == set(model.launch_arguments)
    for key, value in model.launch_arguments.items():
        assert target_launch_arguments[key] == value

    # with extra_information
    extra_information = str(TEST_DATA_DIR / "extra_information.json")
    with open(extra_information, 'r') as json_file:
        target_extra_information = json.load(json_file)
    args = Namespace(model_name=os.environ["MODEL_NAME"], model=os.environ['MODEL'], with_launch_arguments=False, extra_information=extra_information)
    model = init_model()
    await model.loading(happy_vllm_build_async_engine_client(args), args)
    assert model.launch_arguments == {}
    assert set(target_extra_information) == set(model.extra_information)
    for key, value in model.extra_information.items():
        assert target_extra_information[key] == value


def test_tokenize():
    model = init_model()
    prompts = ["Hey it's me. How are you ?", "Fine. You ?", " I'm feeling really great "]
    for prompt in prompts:
        result = model.tokenize(prompt)
        target = list(utils.proper_tokenization(model._tokenizer, prompt))
        assert result == target


def test_split_text():
    model = init_model()
    text = "Hey, my name is LLM. How are you ? Fine, you ? That's wonderful news : I'm also fine. But do you think it will last ?"
        
    # Nominal, all split
    num_tokens_in_chunk = 2
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ?", 
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    split_text = model.split_text(text, num_tokens_in_chunk)
    assert split_text == target_split_text

    # Nominal, some split
    num_tokens_in_chunk = 6
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ? Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    split_text = model.split_text(text, num_tokens_in_chunk)
    assert split_text == target_split_text

    # Nominal, no split
    num_tokens_in_chunk = 1000
    target_split_text = [text]
    split_text = model.split_text(text, num_tokens_in_chunk)
    assert split_text == target_split_text

    # Change separator tokens
    num_tokens_in_chunk = 2
    separators = [" ?"]
    target_split_text = ["Hey, my name is LLM. How are you ?",
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine. But do you think it will last ?"]
    split_text = model.split_text(text, num_tokens_in_chunk, separators=separators)
    assert split_text == target_split_text

    # No separator
    num_tokens_in_chunk = 2
    separators = []
    target_split_text = [text]
    split_text = model.split_text(text, num_tokens_in_chunk, separators=separators)
    assert split_text == target_split_text

    # Separator not present in the text
    num_tokens_in_chunk = 2
    new_text = "Unfortunately, i haven't put any separators in this sentence; I wonder what will happen ;"
    target_split_text = [new_text]
    split_text = model.split_text(new_text, num_tokens_in_chunk, separators=separators)
    assert split_text == target_split_text


def test_extract_text_outside_truncation():
    model = init_model()
    max_length = 4

    prompt = "Hey, it's me. How are you ?"

    # Left truncation
    truncation_side = "left"
    token_ids = model._tokenizer(prompt, add_special_tokens=False)['input_ids']
    target_truncated_prompt = model._tokenizer.decode(token_ids[:-max_length])
    truncated_prompt = model.extract_text_outside_truncation(prompt, truncation_side, max_length)
    assert truncated_prompt == target_truncated_prompt

    # Right truncation
    truncation_side = "right"
    token_ids = model._tokenizer(prompt, add_special_tokens=False)['input_ids']
    target_truncated_prompt = model._tokenizer.decode(token_ids[max_length:])
    truncated_prompt = model.extract_text_outside_truncation(prompt, truncation_side, max_length)
    assert truncated_prompt == target_truncated_prompt

    # No truncation
    max_length = 100
    truncation_side = "left"
    target_truncated_prompt = ""
    truncated_prompt = model.extract_text_outside_truncation(prompt, truncation_side, max_length)
    assert truncated_prompt == target_truncated_prompt


def test_find_indices_sub_list_in_list():
    # Nominal case
    big_list = [3, 4, 1, 2, 3, 4, 5, 6, 3, 4]
    sub_list = [3, 4]
    target_indices = [1, 5, 9]
    assert model_base.find_indices_sub_list_in_list(big_list, sub_list) == target_indices

    # Presence of the beginning of sub_list
    big_list = [3, 4, 1, 2, 3, 4, 5, 6, 3]
    sub_list = [3, 4]
    target_indices = [1, 5]
    assert model_base.find_indices_sub_list_in_list(big_list, sub_list) == target_indices

    big_list = [3, 4, 1, 2, 3, 5, 6, 3, 4]
    sub_list = [3, 4]
    target_indices = [1, 8]
    assert model_base.find_indices_sub_list_in_list(big_list, sub_list) == target_indices

    # Presence the end of sub_list
    big_list = [3, 4, 1, 2, 4, 5, 6, 3, 4]
    sub_list = [3, 4]
    target_indices = [1, 8]
    assert model_base.find_indices_sub_list_in_list(big_list, sub_list) == target_indices

    big_list = [3, 4, 1, 2, 3, 4, 5, 6, 4]
    sub_list = [3, 4]
    target_indices = [1, 5]
    assert model_base.find_indices_sub_list_in_list(big_list, sub_list) == target_indices


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Technical schemas"""
import os
import json
from pydantic import BaseModel, Field

from vllm.entrypoints.openai.protocol import ModelList

# Load the response examples
directory = os.path.dirname(os.path.abspath(__file__))
response_examples_path = os.path.join(directory, "examples", "response.json")
with open(response_examples_path, 'r') as file:
    response_examples = json.load(file)


class ResponseLiveness(BaseModel):
    """Return object for liveness probe"""

    alive: str = Field(None, title="Alive message")
    model_config = {
        "json_schema_extra": {
            "examples": [
                response_examples["liveness"]
            ]
        }
    }

class ResponseReadiness(BaseModel):
    """Return object for readiness probe"""

    ready: str = Field(None, title="Ready message")
    model_config = {
        "json_schema_extra": {
            "examples": [
                response_examples["readiness"]
            ]
        }
    }


class ResponseInformation(BaseModel):
    """Return object for info resource"""

    application: str = Field(None, title="Application name")
    version: str = Field(None, title="Application version")
    vllm_version: str = Field(None, title="Version of vLLM")
    model_name: str = Field(None, title="Model name")
    truncation_side: str = Field(None, title="Truncation side")
    max_length : int = Field(None, title="Max length")
    extra_information: dict = Field(None, title="Extra information")
    model_config = {
        "json_schema_extra": {
            "examples": [
                response_examples["information"]
            ]
        }
    }


class ResponseLiveMetrics(BaseModel):
    requests_running: int = Field(None, title="Number of running requests")
    requests_swapped: int = Field(None, title="Number of swapped requests")
    requests_pending: int = Field(None, title="Number of pending requests")
    gpu_cache_usage: float = Field(None, title="GPU cache usage")
    cpu_cache_usage: float = Field(None, title="CPU cache usage")
    model_config = {
        "json_schema_extra": {
            "examples": [
                response_examples["live_metrics"]
            ]
        }
    }


class HappyvllmModelList(ModelList):
    model_config = {"json_schema_extra": {"examples": [response_examples["models"]]}}





#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from pathlib import Path
import importlib.metadata
from typing import List, Union


logger = logging.getLogger(__name__)

TEST_TOKENIZER_NAME = "mistralai/Mixtral-8x7B-Instruct-v0.1"

def get_package_version() -> str:
    '''Returns the current version of the package

    Returns:
        str: version of the package
    '''
    version = importlib.metadata.version("happy_vllm")
    return version


def get_vllm_version() -> str:
    '''Returns the installed version of vLLM 

    Returns:
        str: version of vLLM
    '''
    version = importlib.metadata.version("vllm")
    return version


def proper_tokenization(tokenizer, str_to_tokenize: str) -> tuple:
    """Gets the token ids for a str. We must use this technique in order to get the right token ids
    with their whitespace in the case of some tokenizers (e.g. Llama) adding whitespace in front of the tokenized sentence

    Args:
        tokenizer : A tokenizer
        str_to_tokenize (str) : The string one wants to tokenize

    Returns:
        tuple : The tuple containing the token ids for the input string
    """
    big_word = 'thisisareallybigwordisntit'
    token_ids_big_word = tokenizer(big_word, add_special_tokens=False)['input_ids']
    # We concatenate the big word with the str in order for the str not to be at the beginning of the sentence
    new_text = big_word + str_to_tokenize
    token_ids_new_text = tokenizer(new_text, add_special_tokens=False)['input_ids']
    # We check that part of the str was not "integrated" in a token of the big_word
    while token_ids_big_word != token_ids_new_text[:len(token_ids_big_word)]:
        # If it has been "integrated", we take another big_word
        big_word = big_word[:-1]
        token_ids_big_word = tokenizer(big_word, add_special_tokens=False)['input_ids']
        # We concatenate the big word with the str in order for the str not to be at the beginning of the sentence
        new_text = big_word + str_to_tokenize
        token_ids_new_text = tokenizer(new_text, add_special_tokens=False)['input_ids']
    token_ids_str = tuple(token_ids_new_text[len(token_ids_big_word):])
    return token_ids_str


def proper_decode(tokenizer, token_ids: Union[int, List[int]]) -> str:
    """Gets the corresponding string from the token ids. We must use this technique in order to get the right string
    with their whitespace in the case of some tokenizers (e.g. Llama) deleting whitespace in front of the tokenized sentence
    """
    if isinstance(token_ids, int):
        token_ids = [token_ids]
    extra_token_id = proper_tokenization(tokenizer, 'c')[0]
    token_ids = [extra_token_id] + token_ids
    return tokenizer.decode(token_ids)[1:]


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from fastapi import FastAPI
from argparse import Namespace
from .routers import main_routeur
from .core.resources import get_lifespan
from prometheus_client import make_asgi_app
from fastapi.middleware.cors import CORSMiddleware
from vllm.entrypoints.openai.api_server import mount_metrics
from vllm.engine.multiprocessing.client import MQLLMEngineClient

from happy_vllm import utils
from happy_vllm.middlewares.exception import ExceptionHandlerMiddleware


async def declare_application(async_engine_client: MQLLMEngineClient, args: Namespace) -> FastAPI:
    """Create the FastAPI application

    See https://fastapi.tiangolo.com/tutorial/first-steps/ to learn how to
    customize your FastAPI application
    """
    if args.disable_fastapi_docs :
        app = FastAPI(
            openapi_url=None,
            docs_url=None,
            redoc_url=None,
            title=f"A REST API for vLLM",
            description=f"A REST API for vLLM, production ready",
            lifespan=get_lifespan(async_engine_client, args=args),
            version=utils.get_package_version()
        )
    else:
        app = FastAPI(
            title=f"A REST API for vLLM",
            description=f"A REST API for vLLM, production ready",
            lifespan=get_lifespan(async_engine_client, args=args),
            version=utils.get_package_version()
        )

    # Add prometheus asgi middleware to route /metrics requests
    mount_metrics(app)

    # CORS middleware that allows all origins to avoid CORS problems
    # see https://fastapi.tiangolo.com/tutorial/cors/#use-corsmiddleware
    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )

    # Add exception middleware
    if args.explicit_errors:
        app.add_middleware(ExceptionHandlerMiddleware)

    #
    app.include_router(main_routeur, prefix=args.api_endpoint_prefix)

    app.root_path = args.root_path

    app.state.enable_server_load_tracking = args.enable_server_load_tracking
    app.state.server_load_metrics = 0

    return app


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import signal
import logging

from vllm.config import VllmConfig
from prometheus_client import Gauge
from vllm.usage.usage_lib import UsageContext
from vllm.engine.multiprocessing.engine import MQLLMEngine


logger = logging.getLogger(__name__)


def run_mp_engine(
        vllm_config: VllmConfig, usage_context: UsageContext,
        ipc_path: str, disable_log_stats: bool,
        disable_log_requests: bool, engine_alive):
    
    try :
        def signal_handler(*_) -> None:
            # Interrupt server on sigterm
            raise KeyboardInterrupt("MQLLMEngine terminated")
        
        engine = MQLLMEngine.from_vllm_config(
            vllm_config=vllm_config,
            usage_context=usage_context,
            disable_log_stats=disable_log_stats,
            disable_log_requests=disable_log_requests,
            ipc_path=ipc_path)
        model_consumed_memory = Gauge("model_memory_usage", "Model Consumed GPU Memory in GB ")
        if vllm_config.scheduler_config.num_scheduler_steps > 1 :
            model_consumed_memory.set(round(engine.engine.model_executor.driver_worker.model_runner._base_model_runner.model_memory_usage/float(2**30),2)) # type: ignore
        else:
            model_consumed_memory.set(round(engine.engine.model_executor.driver_worker.model_runner.model_memory_usage/float(2**30),2)) # type: ignore
        signal.signal(signal.SIGTERM, signal_handler)
        engine.start()
    except BaseException as e:
        logger.exception(e)
        engine_alive.value = False
        raise e
     


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import signal
import asyncio
import logging
import uvicorn
import argparse

from vllm.entrypoints.launcher import serve_http 
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.utils import is_valid_ipv6_address, set_ulimit
import vllm.entrypoints.openai.api_server as vllm_api_server
from vllm.entrypoints.openai.tool_parsers import ToolParserManager
from vllm.entrypoints.openai.cli_args import validate_parsed_serve_args
from vllm.entrypoints.openai.reasoning_parsers import ReasoningParserManager

from happy_vllm.utils_args import parse_args
from happy_vllm.engine.mp_engine import run_mp_engine
from happy_vllm.application import declare_application


TIMEOUT_KEEP_ALIVE = 5 # seconds
logger = logging.getLogger(__name__)


def main(**uvicorn_kwargs) -> None:
    args = parse_args()
    asyncio.run(launch_app(args, **uvicorn_kwargs))
    

def happy_vllm_build_async_engine_client(args):
    """Replace vllm.entrypoints.openai.api_server.run_rpc_server by happy_vllm.run_rpc_server
    """
    vllm_api_server.run_mp_engine  = run_mp_engine
    return vllm_api_server.build_async_engine_client(args)


async def launch_app(args, **uvicorn_kwargs):
    # Check args
    validate_parsed_serve_args(args)
    # Register new tool parser
    if args.tool_parser_plugin and len(args.tool_parser_plugin) > 3:
        ToolParserManager.import_tool_parser(args.tool_parser_plugin)
    valide_tool_parses = ToolParserManager.tool_parsers.keys()
    if args.enable_auto_tool_choice \
        and args.tool_call_parser not in valide_tool_parses:
        raise KeyError(f"invalid tool call parser: {args.tool_call_parser} "
                       f"(chose from {{ {','.join(valide_tool_parses)} }})")
    valid_reasoning_parses = ReasoningParserManager.reasoning_parsers.keys()
    if args.enable_reasoning \
        and args.reasoning_parser not in valid_reasoning_parses:
        raise KeyError(
            f"invalid reasoning parser: {args.reasoning_parser} "
            f"(chose from {{ {','.join(valid_reasoning_parses)} }})")

    # Bind socket
    sock_addr = (args.host or "", args.port)
    sock = vllm_api_server.create_server_socket(sock_addr)

    set_ulimit()

    def signal_handler(*_) -> None:
        # Interrupt server on sigterm while initializing
        raise KeyboardInterrupt("terminated")
    signal.signal(signal.SIGTERM, signal_handler)

    # Launch app
    async with happy_vllm_build_async_engine_client(args) as async_engine_client:
        app = await declare_application(async_engine_client, args=args)

        def _listen_addr(a: str) -> str:
            if is_valid_ipv6_address(a):
                return '[' + a + ']'
            return a or "127.0.0.1"

        is_ssl = args.ssl_keyfile and args.ssl_certfile
        logger.info("Starting vLLM API server on http%s://%s:%d",
                    "s" if is_ssl else "", _listen_addr(sock_addr[0]),
                    sock_addr[1])

        shutdown_task = await serve_http(app,
                                        sock=sock,
                                        enable_ssl_refresh=args.enable_ssl_refresh,
                                        host=args.host,
                                        port=args.port,
                                        log_level=args.uvicorn_log_level,
                                        # NOTE: When the 'disable_uvicorn_access_log' value is True,
                                        # no access log will be output.
                                        access_log=not args.disable_uvicorn_access_log,
                                        timeout_keep_alive=TIMEOUT_KEEP_ALIVE,
                                        ssl_keyfile=args.ssl_keyfile,
                                        ssl_certfile=args.ssl_certfile,
                                        ssl_ca_certs=args.ssl_ca_certs,
                                        ssl_cert_reqs=args.ssl_cert_reqs,
                                        **uvicorn_kwargs)
    await shutdown_task

    sock.close()

    
if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import pytest
import numpy as np

from happy_vllm.routers.schemas.utils import NumpyArrayEncoder


def test_numpy_encoder():
    """Test the NumpyArrayEncoder that is used by default to handle numpy objects"""
    obj = np.array([0.1, 0.2], dtype=np.longdouble)
    assert json.dumps(obj, cls=NumpyArrayEncoder) == "[0.1, 0.2]"

    obj = {0.1, 0.2}
    assert json.dumps(obj, cls=NumpyArrayEncoder) == "[0.1, 0.2]"

    obj = np.array([1, 2], dtype=np.int16)
    assert json.dumps(obj, cls=NumpyArrayEncoder) == "[1, 2]"

    obj = np.array(["a", "b"])
    assert json.dumps(obj, cls=NumpyArrayEncoder) == '["a", "b"]'

    with pytest.raises(TypeError):
        assert json.dumps(str, cls=NumpyArrayEncoder)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import json
import vllm.envs as envs
from vllm.utils import random_uuid
from pydantic import BaseModel, Field
from starlette.requests import Request
from typing_extensions import assert_never
from vllm.sampling_params import SamplingParams
from vllm.entrypoints.utils import with_cancellation, load_aware_call
from vllm.engine.async_llm_engine import AsyncLLMEngine
from lmformatenforcer import TokenEnforcerTokenizerData
from vllm.entrypoints.openai import protocol as vllm_protocol
from vllm.entrypoints.openai.serving_engine import OpenAIServing
from fastapi import APIRouter, Body, Depends, HTTPException, Form
from vllm.entrypoints.openai.api_server import validate_json_request
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from vllm.entrypoints.openai.serving_pooling import OpenAIServingPooling
from starlette.responses import JSONResponse, Response, StreamingResponse
from typing import Annotated, AsyncGenerator, Tuple, List, Optional, Union
from vllm.entrypoints.openai.serving_embedding import OpenAIServingEmbedding
from vllm.entrypoints.openai.serving_tokenization import OpenAIServingTokenization

from happy_vllm import utils

from ..model.model_base import Model
from ..core.resources import RESOURCE_MODEL, RESOURCES
from happy_vllm.routers.schemas import functional as functional_schema


# Load the response examples
directory = os.path.dirname(os.path.abspath(__file__))
request_examples_path = os.path.join(directory, "schemas", "examples", "request.json")
with open(request_examples_path, 'r') as file:
    request_openapi_examples = json.load(file)

# Functional router
router = APIRouter()


def verify_request(request: Union[
    vllm_protocol.ChatCompletionRequest,
    vllm_protocol.CompletionRequest]) -> None:
    """Parses the sampling parameters to check if any combination will break the app

    Args:
        request  (Union[ChatCompletionRequest, CompletionRequest]): The request to verify
    Returns:
        None
    """
    status_code = 422
    detail = None
    if request.echo and request.stream:
        detail="Use both echo and stream breaks backend"
    if request.temperature is not None and request.top_p is not None:
        if request.temperature == 0 and request.top_p == 0:
            detail=f"Use temperature and top_p equal to 0 breaks the model"
    if request.temperature and request.top_k:
        if request.temperature > 2 and request.top_k == 1:
            detail=f"Use temperature with high value: {request.temperature} and top_k equals to 1 : {request.top_k} breaks the model"
    if request.top_p and request.top_k:
        if request.top_p == 1 and request.top_k == 1:
            detail=f"Use top_p and top_k equal to 1 breaks the model"
    if request.max_tokens and request.min_tokens:
        if request.max_tokens < request.min_tokens:
            detail=f"Use max_tokens: {request.max_tokens} less than min_tokens : {request.min_tokens} breaks the model"
    if detail :
        raise HTTPException(
            status_code=status_code, 
            detail=detail
        )


def check_generator(generator: Union[
    vllm_protocol.ChatCompletionResponse, 
    vllm_protocol.CompletionResponse
    ]) -> None:
    """Parses the LLM response to check if prompt_logprobs and fix '-inf' value

    Args:
        request  (Union[ChatCompletionResponse, CompletionResponse]): The request to verify
    Returns:
        None
    """
    if hasattr(generator, "prompt_logprobs"):
        if generator.prompt_logprobs:
            for logprob_dict in generator.prompt_logprobs:
                    if logprob_dict:
                        for logprob_values in logprob_dict.values():
                            if logprob_values.logprob == float('-inf'):
                                logprob_values.logprob = -9999.0


def base(request: Request, model: Model) -> OpenAIServing:
    return model.openai_serving_tokenization


@router.post("/v1/tokenizer", response_model=functional_schema.ResponseTokenizer)
async def tokenizer(request: Request,
    request_type: Annotated[
        functional_schema.RequestTokenizer,
        Body(openapi_examples=request_openapi_examples["tokenizer"])] = None
    ) -> Response:
    """Tokenizes a text

    The request should be a JSON object with the following fields:
    - text: The text to tokenize
    - with_tokens_str (optional): Whether we want the tokens strings in the output
    - vanilla (optional) : Whether we want the vanilla version of the tokenizers
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    request_dict = await request.json()
    text = request_dict.pop("text")
    vanilla = request_dict.get("vanilla", True)
    with_tokens_str = request_dict.get('with_tokens_str', False)
    
    if vanilla:
        tokens_ids = model._tokenizer(text)['input_ids']
        if with_tokens_str:
            tokens_str = model._tokenizer.convert_ids_to_tokens(tokens_ids)
    else:
        tokens_ids = model.tokenize(text)
        if with_tokens_str:
            tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in tokens_ids]
            

    ret = {"tokens_ids": tokens_ids, "tokens_nb": len(tokens_ids)}
    if with_tokens_str:
        ret['tokens_str'] = tokens_str
    return JSONResponse(ret)


@router.post("/v2/tokenizer", response_model=vllm_protocol.TokenizeResponse, dependencies=[Depends(validate_json_request)])
@with_cancellation
async def tokenizer_v2(request: Annotated[vllm_protocol.TokenizeRequest,
        Body(openapi_examples=request_openapi_examples["vllm_tokenizer"])],
        raw_request: Request
    ):
    """Tokenizes a text

    The request should be a JSON object with the following fields:

    Completions :
    - model : ID of the model to use
    - prompt : The text to tokenize
    - add_special_tokens : Add a special tokens to the begin (optional, default value : `true`)
    
    Chat Completions:
    - model : ID of the model to use
    - messages: The texts to tokenize
    - add_special_tokens : Add a special tokens to the begin (optional, default value : `false`)
    - add_generation_prompt : Add generation prompt's model in decode response (optional, default value : `true`)
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    generator = await model.openai_serving_tokenization.create_tokenize(request, raw_request)
    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)
    else:
        if not isinstance(generator, vllm_protocol.TokenizeResponse):
            raise TypeError("Expected generator to be an instance of vllm_protocol.TokenizeResponse")
        return JSONResponse(content=generator.model_dump())
    

@router.post("/v1/decode", response_model=functional_schema.ResponseDecode)
async def decode(request: Request,
    request_type: Annotated[
        functional_schema.RequestDecode,
        Body(openapi_examples=request_openapi_examples["decode"])] = None
    ) -> Response:
    """Decodes token ids

    The request should be a JSON object with the following fields:
    - token_ids: The ids of the tokens
    - with_tokens_str : If the result should also include a list of str
    - vanilla (optional) : Whether we want the vanilla version of the tokenizers
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    request_dict = await request.json()
    token_ids = request_dict.pop("token_ids")
    with_tokens_str = request_dict.get("with_tokens_str", False)
    vanilla = request_dict.get("vanilla", True)
    
    if vanilla:
        decoded_string = model._tokenizer.decode(token_ids)
        if with_tokens_str:
            tokens_str = model._tokenizer.convert_ids_to_tokens(token_ids)
    else:
        decoded_string = utils.proper_decode(model._tokenizer, token_ids)
        if with_tokens_str:
            tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in token_ids]
            

    ret = {"decoded_string": decoded_string}
    if with_tokens_str:
        ret[ "tokens_str"] = tokens_str
    return JSONResponse(ret)


@router.post("/v2/decode", response_model=vllm_protocol.DetokenizeResponse, dependencies=[Depends(validate_json_request)])
@with_cancellation
async def decode_v2(request :Annotated[
        vllm_protocol.DetokenizeRequest,
        Body(openapi_examples=request_openapi_examples["vllm_decode"])],
        raw_request: Request
    ):
    """Decodes token ids

    The request should be a JSON object with the following fields:
    - tokens: The ids of the tokens
    - model : ID of the model to use
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    generator = await model.openai_serving_tokenization.create_detokenize(request, raw_request)
    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)
    else:
        if not isinstance(generator, vllm_protocol.DetokenizeResponse):
            raise TypeError("Expected generator to be an instance of vllm_protocol.DetokenizeResponse")
        return JSONResponse(content=generator.model_dump())


@router.post("/v1/split_text", response_model=functional_schema.ResponseSplitText)
async def split_text(request: Request,
    request_type: Annotated[
        functional_schema.RequestSplitText,
        Body(openapi_examples=request_openapi_examples["split_text"])] = None
    ):
    """Splits a text with a minimal number of token in each chunk. Each chunk is delimited by a separator

    The request should be a JSON object with the following fields:
    - text: The text to split
    - num_tokens_in_chunk (optional): The minimal number of tokens we want in each chunk
    - separators (optional) : The allowed separators between the chunks
    """
    model: Model = RESOURCES[RESOURCE_MODEL]

    request_dict = await request.json()
    split_text = model.split_text(**request_dict)
    response = {"split_text": split_text}

    return JSONResponse(response)


@router.post("/v1/metadata_text", response_model=functional_schema.ResponseMetadata)
async def metadata_text(request: Request,
    request_type: Annotated[
        functional_schema.RequestMetadata,
        Body(openapi_examples=request_openapi_examples["metadata_text"])] = None):
    """Gives meta data on a text

    The request should be a JSON object with the following fields:
    - text: The text to parse
    - truncation_side (optional): The truncation side of the tokenizer
    - max_length (optional) : The max length before truncation

    The default values for truncation_side and max_length are those of the underlying model
    """
    model: Model = RESOURCES[RESOURCE_MODEL]

    request_dict = await request.json()

    tokens_ids = model.tokenize(request_dict['text'])
    truncated_text = model.extract_text_outside_truncation(**request_dict)
    ret = {"tokens_nb": len(tokens_ids), "truncated_text": truncated_text}

    return JSONResponse(ret)


@router.post(
    "/v1/chat/completions", 
    response_model=Union[
        vllm_protocol.ErrorResponse, 
        functional_schema.HappyvllmChatCompletionResponse
    ], 
    dependencies=[Depends(validate_json_request)]
)
@with_cancellation
@load_aware_call
async def create_chat_completion(
    request: Annotated[vllm_protocol.ChatCompletionRequest, 
    Body(openapi_examples=request_openapi_examples["chat_completions"])],
    raw_request: Request
):
    """Open AI compatible chat completion. See https://docs.vllm.ai/en/latest/serving/openai_compatible_server.html for more details
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    handler = model.openai_serving_chat
    if handler is None:
        raise HTTPException(
            status_code=400, 
            detail=f"The model does not support Chat Completions API"
        )
    verify_request(request)
    generator = await handler.create_chat_completion(
        request, raw_request)
    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)
    check_generator(generator)
    if request.stream:
        return StreamingResponse(content=generator, # type: ignore
                                 media_type="text/event-stream")
    else:
        return JSONResponse(content=generator.model_dump()) # type: ignore


@router.post(
    "/v1/completions", 
    response_model=Union[
        vllm_protocol.ErrorResponse, 
        functional_schema.HappyvllmCompletionResponse
    ], 
    dependencies=[Depends(validate_json_request)]
)
@with_cancellation
@load_aware_call
async def create_completion(request: Annotated[vllm_protocol.CompletionRequest, Body(openapi_examples=request_openapi_examples["completions"])],
                            raw_request: Request):
    """Open AI compatible completion. See https://docs.vllm.ai/en/latest/serving/openai_compatible_server.html for more details
    """
    model: Model = RESOURCES[RESOURCE_MODEL]
    handler = model.openai_serving_completion
    if handler is None:
        return base(raw_request, model).create_error_response(
            message="The model does not support Completions API")
    verify_request(request)
    generator = await handler.create_completion(
        request, raw_request)
    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)
    check_generator(generator)
    if request.stream:
        return StreamingResponse(content=generator,
                                 media_type="text/event-stream")
    else:
        return JSONResponse(content=generator.model_dump())


@router.post("/v1/audio/transcriptions")
@with_cancellation
@load_aware_call
async def create_transcriptions(request: Annotated[vllm_protocol.TranscriptionRequest,
                                                   Form()],
                                raw_request: Request):
    model: Model = RESOURCES[RESOURCE_MODEL]
    handler = model.openai_serving_transcription
    if handler is None:
        return base(raw_request, model).create_error_response(
            message="The model does not support Transcriptions API")

    audio_data = await request.file.read()
    generator = await handler.create_transcription(audio_data, request,
                                                   raw_request)

    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)

    elif isinstance(generator, vllm_protocol.TranscriptionResponse):
        return JSONResponse(content=generator.model_dump())

    return StreamingResponse(content=generator, media_type="text/event-stream")


@router.post(
    "/v1/embeddings", 
    response_model=Union[
        vllm_protocol.ErrorResponse, 
        vllm_protocol.EmbeddingResponse
    ], 
    dependencies=[Depends(validate_json_request)]
)
@load_aware_call
async def create_embedding(request: vllm_protocol.EmbeddingRequest, raw_request: Request):
    model: Model = RESOURCES[RESOURCE_MODEL]
    handler = model.openai_serving_embedding
    if handler is None:
        return base(raw_request, model).create_error_response(
                message="The model does not support Embeddings API")
    generator = await handler.create_embedding(request, raw_request)

    if isinstance(generator, vllm_protocol.ErrorResponse):
        return JSONResponse(content=generator.model_dump(),
                            status_code=generator.code)
    elif isinstance(generator, vllm_protocol.EmbeddingResponse):
        return JSONResponse(content=generator.model_dump())

    assert_never(generator)


@router.post("/v1/abort_request")
async def abort_request(request: functional_schema.RequestAbortRequest):
    model: Model = RESOURCES[RESOURCE_MODEL]
    model._model.engine.abort_request(request.request_id)


if envs.VLLM_ALLOW_RUNTIME_LORA_UPDATING:

    @router.post("/v1/load_lora_adapter", dependencies=[Depends(validate_json_request)])
    async def load_lora_adapter(request: vllm_protocol.LoadLoRAAdapterRequest):
        model: Model = RESOURCES[RESOURCE_MODEL]
        response = await model.openai_serving_chat.load_lora_adapter(request)
        if isinstance(response, vllm_protocol.ErrorResponse):
            return JSONResponse(content=response.model_dump(),
                                status_code=response.code)

        response = await model.openai_serving_completion.load_lora_adapter(request)
        if isinstance(response, vllm_protocol.ErrorResponse):
            return JSONResponse(content=response.model_dump(),
                                status_code=response.code)

        return Response(status_code=200, content=response)

    @router.post("/v1/unload_lora_adapter", dependencies=[Depends(validate_json_request)])
    async def unload_lora_adapter(request: vllm_protocol.UnloadLoRAAdapterRequest):
        model: Model = RESOURCES[RESOURCE_MODEL]
        response = await model.openai_serving_chat.unload_lora_adapter(request)
        if isinstance(response, vllm_protocol.ErrorResponse):
            return JSONResponse(content=response.model_dump(),
                                status_code=response.code)

        response = await model.openai_serving_completion.unload_lora_adapter(request)
        if isinstance(response, vllm_protocol.ErrorResponse):
            return JSONResponse(content=response.model_dump(),
                                status_code=response.code)

        return Response(status_code=200, content=response)


    


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functional schemas"""

import os
import json
from starlette.responses import JSONResponse
from pydantic import BaseModel, Field, conint
from typing import Any, List, Union, Optional
from vllm.entrypoints.openai.protocol import ResponseFormat, CompletionResponse, ChatCompletionResponse

from .utils import NumpyArrayEncoder

# Load the response examples
directory = os.path.dirname(os.path.abspath(__file__))
response_examples_path = os.path.join(directory, "examples", "response.json")
with open(response_examples_path, "r") as file:
    response_examples = json.load(file)


class NumpyJSONResponse(JSONResponse):
    def render(self, content: Any) -> bytes:
        return json.dumps(content, cls=NumpyArrayEncoder).encode()


class ResponseTokenizer(BaseModel):
    tokens_ids: List[int] = Field(None, title="List of token ids")
    tokens_nb: int = Field(None, title="Number of tokens")
    tokens_str: List[str] = Field(None, title="List of decoded tokens")
    model_config = {"json_schema_extra": {"examples": [response_examples["tokenizer"]]}}


class RequestTokenizer(BaseModel):
    text: str = Field(None, title="Text to tokenize")
    with_token_str: bool = Field(None, title="Add strings of tokens")
    vanilla: bool = Field(None, title="Use vanilla tokenizer")


class ResponseDecode(BaseModel):
    decoded_string: str = Field(None, title="Decoded text")
    tokens_str: List[str] = Field(None, title="Decoded tokens")
    model_config = {"json_schema_extra": {"examples": [response_examples["decode"]]}}


class RequestDecode(BaseModel):
    token_ids: List[int] = Field(None, title="List of token ids")
    with_token_str: bool = Field(None, title="Add strings of tokens")
    vanilla: bool = Field(None, title="Use vanilla model")


class ResponseSplitText(BaseModel):
    split_text: List[str] = Field(None, title="Split text")
    model_config = {
        "json_schema_extra": {"examples": [response_examples["split_text"]]}
    }


class RequestSplitText(BaseModel):
    text: str = Field(None, title="Input text")
    num_tokens_in_chunk: int = Field(None, title="Min number of tokens per chunk")
    separators: List[str] = Field(None, title="List of separators")


class ResponseMetadata(BaseModel):
    tokens_nb: int = Field(None, title="Number of tokens")
    truncated_text: str = Field(None, title="Truncated text")
    model_config = {
        "json_schema_extra": {"examples": [response_examples["metadata_text"]]}
    }


class RequestMetadata(BaseModel):
    text: str = Field(None, title="Input text")
    truncation_side: str = Field(None, title="Side of truncation")
    max_length: int = Field(None, title="Max length before truncation")


class HappyvllmCompletionResponse(CompletionResponse):
    model_config = {"json_schema_extra": {"examples": [response_examples["completion_response"]]}}


class HappyvllmChatCompletionResponse(ChatCompletionResponse):
    model_config = {"json_schema_extra": {"examples": [response_examples["chat_completion_response"]]}} 


class RequestAbortRequest(BaseModel):
    request_id: str